/*
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This code is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this code; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.  */

/* Copyright (C) 2002 Ian Zimmerman <itz@speakeasy.org> */

static char rcsid[] = "$Id: gimev.c,v 1.14 2002/09/12 02:32:02 itz Exp $ ";

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <paths.h>
#include <assert.h>
#include <stdio.h>
#include <errno.h>

#ifdef MAJOR_IN_MKDEV
#include <sys/mkdev.h>
#endif 

#ifdef MAJOR_IN_SYSMACROS
#include <sys/sysmacros.h>
#endif

#include "gim_public.h"
#include "gimev_options.h"

struct node {
  char *name;
  unsigned flag;
};

/* This should be enough for a long time */
struct node  mod_table[32];

struct node ev_table[] = {
  {"double", gim_double_ev },
  {"down", gim_down_ev },
  {"drag", gim_drag_ev },
  {"move", gim_move_ev },
  {"multi", gim_multi_ev },
  {"raw", gim_raw_ev },
  {"single", gim_single_ev },
  {"triple", gim_triple_ev },
  {"up", gim_up_ev },

  { 0, 0 }
};

void
getmask_one (const char *arg, const struct node* which, unsigned* where)
{
  int mode = 0;                 /* 0 = set, '+' = add, '-' = subtract */
  unsigned value;
  int length;
  const struct node* n;

  if ('+' == arg[0] || '-' == arg[0]) {
    mode = arg[0];
    ++arg;
  }  /*if*/

  if (isdigit(arg[0])) {
    value = atoi(arg);
  } else if (isalpha(arg[0])) {
    length = strspn(arg, "_abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ");
    for (n = which; n->name != 0; ++n) {
      if (length != strlen(n->name)) {
        continue;
      } /*if*/
      if (strncasecmp(n->name, arg, length) == 0) {
        value = n->flag;
        break;
      } /*if*/
    } /*for*/
    if (n->name == 0) print_gimev_docstrings(2);
  } else {
    print_gimev_docstrings(3);
  } /*if*/

  switch (mode) {
  case 0: *where = value; break;
  case '+': *where |= value; break;
  case '-': *where &= ~value; break;
  } /*switch*/
}

void
getmask (const char* arg, const struct node* which, unsigned* where)
{
  const char* p;

  p = arg;
  while (*p != 0) {
    getmask_one(p, which, where);
    if (*p == '-' || *p == '+') p++;
    p += strspn(p, "_abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789");
  } /*while*/
}

int
main (int argc, char** argv)
{
  int vt;
  int status;
  int fd;
  struct stat st;
  struct gim_client_masks masks;
  fd_set readset;
  struct gim_event_t evt;
  char cmd[100];
  char* p;
  char* p2;
  int n;

  parse_gimev_options(argc, argv);

  if (opt_help) print_gimev_docstrings(0);

  /* They can specify the terminal either with an absolute pathname, or a number. */
  /* Anything else is an error. */
  /* The number is what's passed to the server, so if they specify a pathname,
     we stat it and look at its device number. */

  if (opt_terminal != 0) {
    if (isdigit(opt_terminal[0])) {
      vt = atoi(opt_terminal);
    } else if (opt_terminal[0] == '/') {
      status = stat(opt_terminal, &st);
      if (status < 0) {
        perror(opt_terminal);
        exit(1);
      } /*if*/
      if (major(st.st_rdev) != TTY_MAJOR) {
        fprintf(stderr, "%s: not a virtual terminal\n", opt_terminal);
        exit(1);
      } /*if*/
      vt = minor(st.st_rdev);
    } else {
      print_gimev_docstrings(1);
    } /*if*/
  } else {
    vt = -1;
  } /*if*/

  /* The device name is used here to derive the socket pathname.  The default
     device name is "mouse", giving the socket path "/var/run/gim-mouse". */
  if (opt_device == 0) opt_device = "mouse";

  /* Read the modmap file, if available. if not, only numeric mod specs will be available. */
  status = stat(SYSCONFDIR "/gimev_modmap.conf", &st);
  if (status == 0) {
    p = malloc(st.st_size + 1);
    if (p != 0) {
      p[st.st_size] = '\0';
      fd = open(SYSCONFDIR "/gimev_modmap.conf", O_RDONLY);
      if (fd >= 0) {
        status = read(fd, p, st.st_size);
        if (status == st.st_size) {
          p2 = p;
          n = 0;
          while ((p2 < p + st.st_size) && (n < 32)) {
            mod_table[n].name = p2;
            p2 = strchr(p2, '=');
            *p2++ = '\0';
            mod_table[n].flag = strtoul(p2, &p2, 10);
            p2 += strspn(p2, " \t\n");
            n += 1;
          } /*while*/
        } /*if*/
        close(fd);
      } /*if*/
    } /*if*/
  } /*if*/

  /* Slurp the masks */

  masks.minmod = 0;
  if (opt_min_mod != 0) getmask(opt_min_mod, mod_table, &masks.minmod);
  /* this is terrible but I can't be sure the config is available */
  masks.maxmod = ~0;
  if (opt_max_mod != 0) getmask(opt_max_mod, mod_table, &masks.maxmod);
  masks.handle = ~gim_move_ev;
  if (opt_event_mask != 0) getmask(opt_event_mask, ev_table, &masks.handle);
  masks.pass = gim_move_ev;
  if (opt_pass_mask != 0) getmask(opt_pass_mask, ev_table, &masks.pass);

  /* Try to open the connection */

  fd = gim_client_connect(&masks, vt, opt_device);
  if (fd < 0) {
    fprintf(stderr, "%s: %s\n", opt_device, gim_error_string(fd));
    exit(1);
  } /*if*/

  /* Behave nicely in case the output is a pipe */

  setvbuf(stdout, 0, _IOLBF, 0);
  setvbuf(stdin, 0, _IOLBF, 0);

  /* main loop */

  while (1) {
    FD_ZERO(&readset);
    FD_SET(fd, &readset);
    FD_SET(STDIN_FILENO, &readset);
    if (select(fd+1, &readset, 0, 0, 0) < 0 && errno == EINTR) continue;
    if (FD_ISSET(STDIN_FILENO, &readset)) {
      if (fgets(cmd, sizeof(cmd), stdin) == 0) break;
      p = strtok(cmd, " \t\n\r");
      if (p != 0 && strcmp(p, "quit") == 0) {
        break;                  /* out of while */
      } else if (p != 0 && strcmp(p, "help") == 0) {
        print_gimev_help();
      } /*if*/

      status = process_gimev_command(fd, cmd, sizeof(cmd));
      if (status < 0) {
        fprintf(stderr, "%s\n", gim_error_string(status));
        exit(1);
      } else {
        fputs(cmd, stdout); fputc ('\n', stdout);
      } /*if*/

    } /*if*/
    if (FD_ISSET(fd, &readset)) {
      status = gim_get_event(fd, &evt);
      if (status < 0) {
        fprintf(stderr, "%s\n", gim_error_string(status));
        exit(1);
      } /*if*/

      emacsify(&evt);
      if (opt_pointer && (evt.kind & (gim_drag_ev | gim_down_ev)) != 0) {
        status = gim_client_draw_pointer(fd, evt.x, evt.y);
        if (status < 0) {
          fprintf(stderr, "%s\n", gim_error_string(status));
          exit(1);
        } /*if*/
      }
    } /*if*/
  } /*while*/

  /* negative exit codes look like signals, so avoid them */
  return -gim_client_close(fd);
}
